"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AmqpTransporter = exports.AmqpError = void 0;
const amqplib_1 = require("amqplib");
const typescript_microservice_1 = require("typescript-microservice");
const rxjs_1 = require("rxjs");
class AmqpError extends Error {
    code;
    message;
    constructor(code, message) {
        super(message);
        this.code = code;
        this.message = message;
    }
}
exports.AmqpError = AmqpError;
class AmqpTransporter {
    url;
    #push_connection;
    #listen_connection;
    #push_channel;
    #listen_channels;
    #callbacks = [];
    #$on_error = new rxjs_1.Subject();
    constructor(url) {
        this.url = url;
        process.env.DEBUG && this.#$on_error.subscribe(console.error);
    }
    async #listen({ queue_name, topic, options, cb }) {
        const channel = await this.#getListenChannel(options);
        const { queue } = await channel.assertQueue(queue_name, {
            autoDelete: true
        });
        await this.#push_channel.assertExchange(topic, 'topic', {
            autoDelete: true
        });
        await channel.bindQueue(queue, topic, options.route || '#');
        await channel.consume(queue, async (msg) => {
            if (msg == null) {
                this.#$on_error.next('');
                return;
            }
            try {
                channel.ack(msg);
            }
            catch (e) { }
            const { content, properties: { timestamp, messageId, replyTo, } } = msg;
            const data = {
                content,
                created_time: timestamp,
                id: messageId,
                reply_to: replyTo,
                delivery_attempt: msg.properties.headers["x-death"]?.length || 0
            };
            await cb(data);
        }, { noAck: false });
        return channel;
    }
    async #getListenChannel(options) {
        if (options?.limit || this.#listen_channels.length == 0) {
            const channel = await this.#listen_connection.createChannel();
            this.#listen_channels.push(channel);
            options.limit && channel.prefetch(options.limit, false);
            return channel;
        }
        return this.#listen_channels[0];
    }
    async #setup() {
        this.#push_connection = await (0, amqplib_1.connect)(this.url, { reconnectTimeInSeconds: 2, heartbeatIntervalInSeconds: 1 });
        this.#listen_connection = await (0, amqplib_1.connect)(this.url, { reconnectTimeInSeconds: 2, heartbeatIntervalInSeconds: 1 });
        this.#push_channel = await this.#push_connection.createChannel();
        this.#listen_channels = [];
        const subscriptions = [];
        for (const event of ['error', 'close']) {
            subscriptions.push((0, rxjs_1.fromEvent)(this.#push_connection, event).subscribe(this.#$on_error));
            subscriptions.push((0, rxjs_1.fromEvent)(this.#listen_connection, event).subscribe(this.#$on_error));
            subscriptions.push((0, rxjs_1.fromEvent)(this.#push_channel, event).subscribe(this.#$on_error));
        }
        for (const cb of this.#callbacks) {
            const channel = await this.#listen(cb);
            for (const event of ['error', 'close']) {
                subscriptions.push((0, rxjs_1.fromEvent)(channel, event).subscribe(this.#$on_error));
            }
        }
        return {
            clear: async () => {
                await subscriptions.map(s => s.unsubscribe());
                try {
                    await this.#listen_connection.close();
                }
                catch (e) { }
                try {
                    await this.#push_connection.close();
                }
                catch (e) { }
            }
        };
    }
    static async init(url = process.env.AMQP_TRANSPORTER) {
        const amqp = new this(url);
        await new Promise(async (success) => {
            while (true) {
                try {
                    const loop = await amqp.#setup();
                    success(1);
                    await (0, rxjs_1.firstValueFrom)(amqp.#$on_error);
                    await loop.clear();
                }
                catch (e) { }
                console.error(`Error with rabbitmq, reconnect in 1s...`);
                await (0, typescript_microservice_1.sleep)(1000);
            }
        });
        return amqp;
    }
    async publish(topic, data, options = {}) {
        try {
            await this.#push_channel.assertExchange(topic, 'topic', { autoDelete: true, });
            await this.#push_channel.publish(topic, options.route, data, {
                replyTo: options.reply_to,
                messageId: options.id,
            });
        }
        catch (e) {
            process.env.DEBUG && console.error(e);
        }
    }
    listen(topic, cb, options = {}) {
        const queue_name = options.fanout ? '' : `${topic}${options.route || ''}`;
        const config = { cb, options, queue_name, topic };
        this.#callbacks.push(config);
        this.#listen(config);
        return {
            unsubscribe: () => { }
        };
    }
}
exports.AmqpTransporter = AmqpTransporter;
